package org.broadinstitute.dropseqrna.annotation.functionaldata.disambiguate;

public enum FunctionCategory {
    SENSE_CODING,
    RESOLVED_ANTISENSE_CODING,
    ANTISENSE_CODING,
    RESOLVED_SENSE_INTRONIC,
    SENSE_INTRONIC,
    AMBIGUOUS,
    UNRESOLVED,
    UNAMBIGUOUS;
}
