/*
 * MIT License
 *
 * Copyright 2020 Broad Institute
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.broadinstitute.dropseqrna.metrics;

import java.util.Objects;

public class RnaSeqMetricsKey {
    private String sample;
    private String library;
    private String readGroup;

    public RnaSeqMetricsKey(String sample, String library, String readGroup) {
        this.sample = sample;
        this.library = library;
        this.readGroup = readGroup;
    }

    @Override
    public boolean equals(final Object otherObject) {
        if (this == otherObject)
            return true;
        if (!(otherObject instanceof RnaSeqMetricsKey))
            return false;

        final RnaSeqMetricsKey otherKey = (RnaSeqMetricsKey) otherObject;
        return
            Objects.equals(this.sample, otherKey.sample) &&
            Objects.equals(this.library, otherKey.library) &&
            Objects.equals(this.readGroup, otherKey.readGroup);
    }

    public int hashCode() {
        return Objects.hash(sample, library, readGroup);
    }
}
