package org.broadinstitute.dropseqrna.barnyard.digitalallelecounts.sampleassignment;

import org.testng.annotations.Test;

public class GenerateSyntheticDoubletsTest {

	@Test(enabled=false)
	public void doWorkTest() {
		throw new RuntimeException("Test not implemented");
	}
}
