/*
 * MIT License
 *
 * Copyright 2017 Broad Institute
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */
package org.broadinstitute.dropseqrna.annotation;

import org.testng.Assert;
import org.testng.annotations.Test;

import java.util.List;

/**
 * Given a sequence, find a list of GQuadruplex intervals.
 * @author nemesh
 *
 */
public class FindGQuadruplexTest {


	@Test
	public void findSimple () {
		String t1 = "GGGCCTGGGGCTGGGCCTGGG";
		List<GQuadruplex> r1 = GQuadruplex.find("t1", t1);
		Assert.assertEquals(1, r1.size());
		GQuadruplex result = r1.get(0);
		Assert.assertEquals("GGGCCTGGGGCTGGGCCTGGG", result.getSequence());

		Assert.assertEquals(1, result.getMatchInterval().getStart());
		Assert.assertEquals(21, result.getMatchInterval().getEnd());

	}

	@Test
	public void findTandem () {
		String tandem="GGGCCTGGGCCTGGGCCTGGGCCTGGGCCTGGGCCTGGGCCTGGG";
		List<GQuadruplex> r1 = GQuadruplex.find("tandem", tandem);

		Assert.assertEquals(2, r1.size());
		GQuadruplex result = r1.get(0);
		Assert.assertEquals("GGGCCTGGGCCTGGGCCTGGG", result.getSequence());
		Assert.assertEquals(1, result.getMatchInterval().getStart());
		Assert.assertEquals(21, result.getMatchInterval().getEnd());

		result = r1.get(1);
		Assert.assertEquals("GGGCCTGGGCCTGGGCCTGGG", result.getSequence());
		Assert.assertEquals(25, result.getMatchInterval().getStart());
		Assert.assertEquals(45, result.getMatchInterval().getEnd());

	}
}
