/*
 * MIT License
 *
 * Copyright 2018 Broad Institute
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

package org.broadinstitute.dropseqrna.beadsynthesis;

import htsjdk.samtools.metrics.MetricBase;
import htsjdk.samtools.util.Histogram;

public class BeadSynthesisErrorsSummaryMetric extends MetricBase {
	public int NUM_BEADS;
	public int NO_ERROR;
	// public int LOW_UMI_COUNT;
	public int SYNTHESIS_MISSING_BASE;
	public int SINGLE_UMI_ERROR;
	public int PRIMER_MATCH;
	public int FIXED_FIRST_BASE;
	public int OTHER_ERROR_COUNT;


	/** The distribution of  SYNTHESIS_MISSING_BASE error positions */
	private Histogram <Integer> histogram = null;

	public BeadSynthesisErrorsSummaryMetric () {
		this.NUM_BEADS=0;
		this.NO_ERROR=0;
		this.SYNTHESIS_MISSING_BASE=0;
		this.SINGLE_UMI_ERROR=0;
		this.PRIMER_MATCH=0;
		this.OTHER_ERROR_COUNT=0;
		// this.LOW_UMI_COUNT=0;
		histogram = new Histogram<>("SYNTHESIS_ERROR_BASE", "num cells");
	}

	public void incrementSynthesisMissingBase (final int position) {
		histogram.increment(position);
	}

	public Histogram<Integer> getHistogram() {
		return histogram;
	}

}